@@ -7,19 +7,24 @@ import cgi
 import datetime
 import functools
 import inspect
+import netrc
 import os
 import re
 import sys
 import time
-import warnings
 import weakref
 from collections import namedtuple
+from collections.abc import Mapping as ABCMapping
+from contextlib import suppress
 from math import ceil
 from pathlib import Path
+from typing import Any, Dict, List, Optional, Tuple  # noqa
 from urllib.parse import quote
 from urllib.request import getproxies
 
-from async_timeout import timeout
+import async_timeout
+import attr
+from multidict import MultiDict
 from yarl import URL
 
 from . import hdrs
@@ -27,27 +32,25 @@ from .abc import AbstractAccessLogger
 from .log import client_logger
 
 
-try:
-    from asyncio import ensure_future
-except ImportError:
-    ensure_future = getattr(asyncio, 'async')
+__all__ = ('BasicAuth', 'ChainMapProxy')
 
-PY_34 = sys.version_info < (3, 5)
-PY_35 = sys.version_info >= (3, 5)
-PY_352 = sys.version_info >= (3, 5, 2)
+PY_36 = sys.version_info >= (3, 6)
+PY_37 = sys.version_info >= (3, 7)
 
-if sys.version_info >= (3, 4, 3):
-    from http.cookies import SimpleCookie  # noqa
-else:
-    from .backport_cookies import SimpleCookie  # noqa
+if not PY_37:
+    import idna_ssl
+    idna_ssl.patch_match_hostname()
 
 
-__all__ = ('BasicAuth', 'Timeout')
+sentinel = object()  # type: Any
+NO_EXTENSIONS = bool(os.environ.get('AIOHTTP_NO_EXTENSIONS'))  # type: bool
 
+# N.B. sys.flags.dev_mode is available on Python 3.7+, use getattr
+# for compatibility with older versions
+DEBUG = (getattr(sys.flags, 'dev_mode', False) or
+         (not sys.flags.ignore_environment and
+          bool(os.environ.get('PYTHONASYNCIODEBUG'))))
 
-sentinel = object()
-Timeout = timeout
-NO_EXTENSIONS = bool(os.environ.get('AIOHTTP_NO_EXTENSIONS'))
 
 CHAR = set(chr(i) for i in range(0, 128))
 CTL = set(chr(i) for i in range(0, 32)) | {chr(127), }
@@ -56,99 +59,11 @@ SEPARATORS = {'(', ')', '<', '>', '@', ',', ';', ':', '\\', '"', '/', '[', ']',
 TOKEN = CHAR ^ CTL ^ SEPARATORS
 
 
-if PY_35:
-    from collections.abc import Coroutine
-    base = Coroutine
-else:
-    base = object
-
-
-class _BaseCoroMixin(base):
-
-    __slots__ = ('_coro')
-
-    def __init__(self, coro):
-        self._coro = coro
-
-    def send(self, arg):
-        return self._coro.send(arg)
-
-    def throw(self, arg):
-        return self._coro.throw(arg)
-
-    def close(self):
-        return self._coro.close()
-
-    @property
-    def gi_frame(self):
-        return self._coro.gi_frame
-
-    @property
-    def gi_running(self):
-        return self._coro.gi_running
-
-    @property
-    def gi_code(self):
-        return self._coro.gi_code
-
-    def __next__(self):
-        return self.send(None)
-
-    @asyncio.coroutine
-    def __iter__(self):
-        ret = yield from self._coro
-        return ret
-
-    if PY_35:
-        def __await__(self):
-            ret = yield from self._coro
-            return ret
-
-
-if not PY_35:
-    try:
-        from asyncio import coroutines
-        coroutines._COROUTINE_TYPES += (_BaseCoroMixin,)
-    except:  # pragma: no cover
-        pass  # Python 3.4.2 and 3.4.3 has no coroutines._COROUTINE_TYPES
-
-
-class _CoroGuard(_BaseCoroMixin):
-    """Only to be used with func:`deprecated_noop`.
-
-    Otherwise the stack information in the raised warning doesn't line up with
-    the user's code anymore.
-    """
-    __slots__ = ('_msg', '_awaited')
-
-    def __init__(self, coro, msg):
-        super().__init__(coro)
-        self._msg = msg
-        self._awaited = False
-
-    def send(self, arg):
-        self._awaited = True
-        return self._coro.send(arg)
-
-    @asyncio.coroutine
-    def __iter__(self):
-        self._awaited = True
-        return super().__iter__()
-
-    if PY_35:
-        def __await__(self):
-            self._awaited = True
-            return super().__await__()
-
-    def __del__(self):
-        self._coro = None
-        if not self._awaited:
-            warnings.warn(self._msg, DeprecationWarning, stacklevel=2)
-
-
 coroutines = asyncio.coroutines
-old_debug = coroutines._DEBUG
-coroutines._DEBUG = False
+old_debug = coroutines._DEBUG  # type: ignore
+
+# prevent "coroutine noop was never awaited" warning.
+coroutines._DEBUG = False  # type: ignore
 
 
 @asyncio.coroutine
@@ -156,24 +71,15 @@ def noop(*args, **kwargs):
     return
 
 
-def deprecated_noop(message):
-    return _CoroGuard(noop(), message)
-
-
-coroutines._DEBUG = old_debug
-
-
-try:
-    from asyncio import isfuture
-except ImportError:
-    def isfuture(fut):
-        return isinstance(fut, asyncio.Future)
+coroutines._DEBUG = old_debug  # type: ignore
 
 
 class BasicAuth(namedtuple('BasicAuth', ['login', 'password', 'encoding'])):
     """Http basic authentication helper."""
 
-    def __new__(cls, login, password='', encoding='latin1'):
+    def __new__(cls, login: str,
+                password: str='',
+                encoding: str='latin1') -> 'BasicAuth':
         if login is None:
             raise ValueError('None is not allowed as login value')
 
@@ -187,7 +93,7 @@ class BasicAuth(namedtuple('BasicAuth', ['login', 'password', 'encoding'])):
         return super().__new__(cls, login, password, encoding)
 
     @classmethod
-    def decode(cls, auth_header, encoding='latin1'):
+    def decode(cls, auth_header: str, encoding: str='latin1') -> 'BasicAuth':
         """Create a BasicAuth object from an Authorization HTTP header."""
         split = auth_header.strip().split(' ')
         if len(split) == 2:
@@ -207,7 +113,8 @@ class BasicAuth(namedtuple('BasicAuth', ['login', 'password', 'encoding'])):
         return cls(username, password, encoding=encoding)
 
     @classmethod
-    def from_url(cls, url, *, encoding='latin1'):
+    def from_url(cls, url: URL,
+                 *, encoding: str='latin1') -> Optional['BasicAuth']:
         """Create BasicAuth from url."""
         if not isinstance(url, URL):
             raise TypeError("url should be yarl.URL instance")
@@ -215,13 +122,13 @@ class BasicAuth(namedtuple('BasicAuth', ['login', 'password', 'encoding'])):
             return None
         return cls(url.user, url.password or '', encoding=encoding)
 
-    def encode(self):
+    def encode(self) -> str:
         """Encode credentials."""
         creds = ('%s:%s' % (self.login, self.password)).encode(self.encoding)
         return 'Basic %s' % base64.b64encode(creds).decode(self.encoding)
 
 
-def strip_auth_from_url(url):
+def strip_auth_from_url(url: URL) -> Tuple[URL, Optional[BasicAuth]]:
     auth = BasicAuth.from_url(url)
     if auth is None:
         return url, None
@@ -229,12 +136,43 @@ def strip_auth_from_url(url):
         return url.with_user(None), auth
 
 
-ProxyInfo = namedtuple('ProxyInfo', 'proxy proxy_auth')
+def netrc_from_env():
+    netrc_obj = None
+    netrc_path = os.environ.get('NETRC')
+    try:
+        if netrc_path is not None:
+            netrc_path = Path(netrc_path)
+        else:
+            home_dir = Path.home()
+            if os.name == 'nt':  # pragma: no cover
+                netrc_path = home_dir.joinpath('_netrc')
+            else:
+                netrc_path = home_dir.joinpath('.netrc')
+
+        if netrc_path and netrc_path.is_file():
+            try:
+                netrc_obj = netrc.netrc(str(netrc_path))
+            except (netrc.NetrcParseError, OSError) as e:
+                client_logger.warning(".netrc file parses fail: %s", e)
+
+        if netrc_obj is None:
+            client_logger.warning("could't find .netrc file")
+    except RuntimeError as e:  # pragma: no cover
+        """ handle error raised by pathlib """
+        client_logger.warning("could't find .netrc file: %s", e)
+    return netrc_obj
+
+
+@attr.s(frozen=True, slots=True)
+class ProxyInfo:
+    proxy = attr.ib(type=str)
+    proxy_auth = attr.ib(type=BasicAuth)
 
 
 def proxies_from_env():
     proxy_urls = {k: URL(v) for k, v in getproxies().items()
                   if k in ('http', 'https')}
+    netrc_obj = netrc_from_env()
     stripped = {k: strip_auth_from_url(v) for k, v in proxy_urls.items()}
     ret = {}
     for proto, val in stripped.items():
@@ -243,30 +181,24 @@ def proxies_from_env():
             client_logger.warning(
                 "HTTPS proxies %s are not supported, ignoring", proxy)
             continue
+        if netrc_obj and auth is None:
+            auth_from_netrc = netrc_obj.authenticators(proxy.host)
+            if auth_from_netrc is not None:
+                # auth_from_netrc is a (`user`, `account`, `password`) tuple,
+                # `user` and `account` both can be username,
+                # if `user` is None, use `account`
+                *logins, password = auth_from_netrc
+                auth = BasicAuth(logins[0] if logins[0] else logins[-1],
+                                 password)
         ret[proto] = ProxyInfo(proxy, auth)
     return ret
 
 
-if PY_352:
-    def create_future(loop):
-        return loop.create_future()
-else:
-    def create_future(loop):  # pragma: no cover
-        """Compatibility wrapper for the loop.create_future() call introduced in
-        3.5.2."""
-        return asyncio.Future(loop=loop)
-
-
 def current_task(loop=None):
-    if loop is None:
-        loop = asyncio.get_event_loop()
-
-    task = asyncio.Task.current_task(loop=loop)
-    if task is None:
-        if hasattr(loop, 'current_task'):
-            task = loop.current_task()
-
-    return task
+    if PY_37:
+        return asyncio.current_task(loop=loop)
+    else:
+        return asyncio.Task.current_task(loop=loop)
 
 
 def isasyncgenfunction(obj):
@@ -275,21 +207,30 @@ def isasyncgenfunction(obj):
     return False
 
 
+@attr.s(frozen=True, slots=True)
+class MimeType:
+    type = attr.ib(type=str)
+    subtype = attr.ib(type=str)
+    suffix = attr.ib(type=str)
+    parameters = attr.ib(type=MultiDict)
+
+
 def parse_mimetype(mimetype):
     """Parses a MIME type into its components.
 
     mimetype is a MIME type string.
 
-    Returns a 4 element tuple for MIME type, subtype, suffix and parameters.
+    Returns a MimeType object.
 
     Example:
 
     >>> parse_mimetype('text/html; charset=utf-8')
-    ('text', 'html', '', {'charset': 'utf-8'})
+    MimeType(type='text', subtype='html', suffix='',
+             parameters={'charset': 'utf-8'})
 
     """
     if not mimetype:
-        return '', '', '', {}
+        return MimeType(type='', subtype='', suffix='', parameters={})
 
     parts = mimetype.split(';')
     params = []
@@ -298,7 +239,7 @@ def parse_mimetype(mimetype):
             continue
         key, value = item.split('=', 1) if '=' in item else (item, '')
         params.append((key.lower().strip(), value.strip(' "')))
-    params = dict(params)
+    params = MultiDict(params)
 
     fulltype = parts[0].strip().lower()
     if fulltype == '*':
@@ -308,7 +249,8 @@ def parse_mimetype(mimetype):
         if '/' in fulltype else (fulltype, '')
     stype, suffix = stype.split('+', 1) if '+' in stype else (stype, '')
 
-    return mtype, stype, suffix, params
+    return MimeType(type=mtype, subtype=stype, suffix=suffix,
+                    parameters=params)
 
 
 def guess_filename(obj, default=None):
@@ -346,6 +288,9 @@ def content_disposition_header(disptype, quote_fields=True, **params):
     return value
 
 
+KeyMethod = namedtuple('KeyMethod', 'key method')
+
+
 class AccessLogger(AbstractAccessLogger):
     """Helper object to log access.
 
@@ -374,7 +319,7 @@ class AccessLogger(AbstractAccessLogger):
     """
     LOG_FORMAT_MAP = {
         'a': 'remote_address',
-        't': 'request_time',
+        't': 'request_start_time',
         'P': 'process_id',
         'r': 'first_request_line',
         's': 'response_status',
@@ -386,12 +331,10 @@ class AccessLogger(AbstractAccessLogger):
         'o': 'response_header',
     }
 
-    LOG_FORMAT = '%a %l %u %t "%r" %s %b "%{Referrer}i" "%{User-Agent}i"'
+    LOG_FORMAT = '%a %t "%r" %s %b "%{Referer}i" "%{User-Agent}i"'
     FORMAT_RE = re.compile(r'%(\{([A-Za-z0-9\-_]+)\}([ioe])|[atPrsbOD]|Tf?)')
     CLEANUP_RE = re.compile(r'(%[^s])')
-    _FORMAT_CACHE = {}
-
-    KeyMethod = namedtuple('KeyMethod', 'key method')
+    _FORMAT_CACHE = {}  # type: Dict[str, Tuple[str, List[KeyMethod]]]
 
     def __init__(self, logger, log_format=LOG_FORMAT):
         """Initialise the logger.
@@ -414,7 +357,7 @@ class AccessLogger(AbstractAccessLogger):
 
         All known atoms will be replaced with %s
         Also methods for formatting of those atoms will be added to
-        _methods in apropriate order
+        _methods in appropriate order
 
         For example we have log_format = "%a %t"
         This format will be translated to "%s %s"
@@ -430,10 +373,6 @@ class AccessLogger(AbstractAccessLogger):
         also receive key name (by functools.partial)
 
         """
-
-        log_format = log_format.replace("%l", "-")
-        log_format = log_format.replace("%u", "-")
-
         # list of (key, method) tuples, we don't use an OrderedDict as users
         # can repeat the same key more than once
         methods = list()
@@ -447,7 +386,7 @@ class AccessLogger(AbstractAccessLogger):
                 m = getattr(AccessLogger, '_format_%s' % atom[2])
                 m = functools.partial(m, atom[1])
 
-            methods.append(self.KeyMethod(format_key, m))
+            methods.append(KeyMethod(format_key, m))
 
         log_format = self.FORMAT_RE.sub(r'%s', log_format)
         log_format = self.CLEANUP_RE.sub(r'%\1', log_format)
@@ -475,7 +414,9 @@ class AccessLogger(AbstractAccessLogger):
 
     @staticmethod
     def _format_t(request, response, time):
-        return datetime.datetime.utcnow().strftime('[%d/%b/%Y:%H:%M:%S +0000]')
+        now = datetime.datetime.utcnow()
+        start_time = now - datetime.timedelta(seconds=time)
+        return start_time.strftime('[%d/%b/%Y:%H:%M:%S +0000]')
 
     @staticmethod
     def _format_P(request, response, time):
@@ -524,7 +465,10 @@ class AccessLogger(AbstractAccessLogger):
                 if key.__class__ is str:
                     extra[key] = value
                 else:
-                    extra[key[0]] = {key[1]: value}
+                    k1, k2 = key
+                    dct = extra.get(k1, {})
+                    dct[k2] = value
+                    extra[k1] = dct
 
             self.logger.info(self._log_format % tuple(values), extra=extra)
         except Exception:
@@ -542,13 +486,10 @@ class reify:
 
     def __init__(self, wrapped):
         self.wrapped = wrapped
-        try:
-            self.__doc__ = wrapped.__doc__
-        except:  # pragma: no cover
-            self.__doc__ = ""
+        self.__doc__ = wrapped.__doc__
         self.name = wrapped.__name__
 
-    def __get__(self, inst, owner, _sentinel=sentinel):
+    def __get__(self, inst, owner):
         try:
             try:
                 return inst._cache[self.name]
@@ -565,6 +506,15 @@ class reify:
         raise AttributeError("reified property is read-only")
 
 
+reify_py = reify
+
+try:
+    from ._helpers import reify as reify_c
+    if not NO_EXTENSIONS:
+        reify = reify_c  # type: ignore
+except ImportError:
+    pass
+
 _ipv4_pattern = (r'^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}'
                  r'(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$')
 _ipv6_pattern = (
@@ -630,10 +580,8 @@ def _weakref_handle(info):
     ref, name = info
     ob = ref()
     if ob is not None:
-        try:
+        with suppress(Exception):
             getattr(ob, name)()
-        except:
-            pass
 
 
 def weakref_handle(ob, name, timeout, loop, ceil_timeout=True):
@@ -680,10 +628,8 @@ class TimeoutHandle:
 
     def __call__(self):
         for cb, args, kwargs in self._callbacks:
-            try:
+            with suppress(Exception):
                 cb(*args, **kwargs)
-            except:
-                pass
 
         self._callbacks.clear()
 
@@ -734,7 +680,7 @@ class TimerContext:
             self._cancelled = True
 
 
-class CeilTimeout(Timeout):
+class CeilTimeout(async_timeout.timeout):
 
     def __enter__(self):
         if self._timeout is not None:
@@ -749,6 +695,9 @@ class CeilTimeout(Timeout):
 
 class HeadersMixin:
 
+    ATTRS = frozenset([
+        '_content_type', '_content_dict', '_stored_content_type'])
+
     _content_type = None
     _content_dict = None
     _stored_content_type = sentinel
@@ -763,26 +712,78 @@ class HeadersMixin:
             self._content_type, self._content_dict = cgi.parse_header(raw)
 
     @property
-    def content_type(self, *, _CONTENT_TYPE=hdrs.CONTENT_TYPE):
+    def content_type(self):
         """The value of content part for Content-Type HTTP header."""
-        raw = self._headers.get(_CONTENT_TYPE)
+        raw = self._headers.get(hdrs.CONTENT_TYPE)
         if self._stored_content_type != raw:
             self._parse_content_type(raw)
         return self._content_type
 
     @property
-    def charset(self, *, _CONTENT_TYPE=hdrs.CONTENT_TYPE):
+    def charset(self):
         """The value of charset part for Content-Type HTTP header."""
-        raw = self._headers.get(_CONTENT_TYPE)
+        raw = self._headers.get(hdrs.CONTENT_TYPE)
         if self._stored_content_type != raw:
             self._parse_content_type(raw)
         return self._content_dict.get('charset')
 
     @property
-    def content_length(self, *, _CONTENT_LENGTH=hdrs.CONTENT_LENGTH):
+    def content_length(self):
         """The value of Content-Length HTTP header."""
-        l = self._headers.get(_CONTENT_LENGTH)
-        if l is None:
-            return None
-        else:
-            return int(l)
+        content_length = self._headers.get(hdrs.CONTENT_LENGTH)
+
+        if content_length:
+            return int(content_length)
+
+
+def set_result(fut, result):
+    if not fut.done():
+        fut.set_result(result)
+
+
+def set_exception(fut, exc):
+    if not fut.done():
+        fut.set_exception(exc)
+
+
+class ChainMapProxy(ABCMapping):
+    __slots__ = ('_maps',)
+
+    def __init__(self, maps):
+        self._maps = tuple(maps)
+
+    def __init_subclass__(cls):
+        raise TypeError("Inheritance class {} from ChainMapProxy "
+                        "is forbidden".format(cls.__name__))
+
+    def __getitem__(self, key):
+        for mapping in self._maps:
+            try:
+                return mapping[key]
+            except KeyError:
+                pass
+        raise KeyError(key)
+
+    def get(self, key, default=None):
+        return self[key] if key in self else default
+
+    def __len__(self):
+        # reuses stored hash values if possible
+        return len(set().union(*self._maps))
+
+    def __iter__(self):
+        d = {}
+        for mapping in reversed(self._maps):
+            # reuses stored hash values if possible
+            d.update(mapping)
+        return iter(d)
+
+    def __contains__(self, key):
+        return any(key in m for m in self._maps)
+
+    def __bool__(self):
+        return any(self._maps)
+
+    def __repr__(self):
+        content = ", ".join(map(repr, self._maps))
+        return 'ChainMapProxy({})'.format(content)
