@@ -1,9 +1,6 @@
 """WebSocket protocol versions 13 and 8."""
 
-import base64
-import binascii
 import collections
-import hashlib
 import json
 import random
 import re
@@ -12,15 +9,13 @@ import zlib
 from enum import IntEnum
 from struct import Struct
 
-from . import hdrs
-from .helpers import NO_EXTENSIONS, noop
-from .http_exceptions import HttpBadRequest, HttpProcessingError
+from .helpers import NO_EXTENSIONS
 from .log import ws_logger
 
 
 __all__ = ('WS_CLOSED_MESSAGE', 'WS_CLOSING_MESSAGE', 'WS_KEY',
-           'WebSocketReader', 'WebSocketWriter', 'do_handshake',
-           'WSMessage', 'WebSocketError', 'WSMsgType', 'WSCloseCode')
+           'WebSocketReader', 'WebSocketWriter', 'WSMessage',
+           'WebSocketError', 'WSMsgType', 'WSCloseCode')
 
 
 class WSCloseCode(IntEnum):
@@ -91,10 +86,6 @@ class WSMessage(_WSMessageBase):
         """
         return loads(self.data)
 
-    @property
-    def tp(self):
-        return self.type
-
 
 WS_CLOSED_MESSAGE = WSMessage(WSMsgType.CLOSED, None, None)
 WS_CLOSING_MESSAGE = WSMessage(WSMsgType.CLOSING, None, None)
@@ -147,7 +138,7 @@ if NO_EXTENSIONS:
     _websocket_mask = _websocket_mask_python
 else:
     try:
-        from ._websocket import _websocket_mask_cython
+        from ._websocket import _websocket_mask_cython  # type: ignore
         _websocket_mask = _websocket_mask_cython
     except ImportError:  # pragma: no cover
         _websocket_mask = _websocket_mask_python
@@ -245,11 +236,12 @@ class WSParserState(IntEnum):
 
 class WebSocketReader:
 
-    def __init__(self, queue, compress=True):
+    def __init__(self, queue, max_msg_size, compress=True):
         self.queue = queue
+        self._max_msg_size = max_msg_size
 
         self._exc = None
-        self._partial = []
+        self._partial = bytearray()
         self._state = WSParserState.READ_HEADER
 
         self._opcode = None
@@ -328,7 +320,13 @@ class WebSocketReader:
                     # got partial frame payload
                     if opcode != WSMsgType.CONTINUATION:
                         self._opcode = opcode
-                    self._partial.append(payload)
+                    self._partial.extend(payload)
+                    if (self._max_msg_size and
+                            len(self._partial) >= self._max_msg_size):
+                        raise WebSocketError(
+                            WSCloseCode.MESSAGE_TOO_BIG,
+                            "Message size {} exceeds limit {}".format(
+                                len(self._partial), self._max_msg_size))
                 else:
                     # previous frame was non finished
                     # we should get continuation opcode
@@ -343,15 +341,29 @@ class WebSocketReader:
                         opcode = self._opcode
                         self._opcode = None
 
-                    self._partial.append(payload)
-
-                    payload_merged = b''.join(self._partial)
+                    self._partial.extend(payload)
+                    if (self._max_msg_size and
+                            len(self._partial) >= self._max_msg_size):
+                        raise WebSocketError(
+                            WSCloseCode.MESSAGE_TOO_BIG,
+                            "Message size {} exceeds limit {}".format(
+                                len(self._partial), self._max_msg_size))
 
                     # Decompress process must to be done after all packets
                     # received.
                     if compressed:
+                        self._partial.extend(_WS_DEFLATE_TRAILING)
                         payload_merged = self._decompressobj.decompress(
-                            payload_merged + _WS_DEFLATE_TRAILING)
+                            self._partial, self._max_msg_size)
+                        if self._decompressobj.unconsumed_tail:
+                            left = len(self._decompressobj.unconsumed_tail)
+                            raise WebSocketError(
+                                WSCloseCode.MESSAGE_TOO_BIG,
+                                "Decompressed message size exceeds limit {}".
+                                format(self._max_msg_size + left,
+                                       self._max_msg_size))
+                    else:
+                        payload_merged = bytes(self._partial)
 
                     self._partial.clear()
 
@@ -415,7 +427,7 @@ class WebSocketReader:
                             'Received fragmented control frame')
 
                     has_mask = (second_byte >> 7) & 1
-                    length = (second_byte) & 0x7f
+                    length = second_byte & 0x7f
 
                     # Control frames MUST have a payload
                     # length of 125 bytes or less
@@ -522,11 +534,11 @@ class WebSocketReader:
 
 class WebSocketWriter:
 
-    def __init__(self, stream, *,
+    def __init__(self, protocol, transport, *,
                  use_mask=False, limit=DEFAULT_LIMIT, random=random.Random(),
                  compress=0, notakeover=False):
-        self.stream = stream
-        self.writer = stream.transport
+        self.protocol = protocol
+        self.transport = transport
         self.use_mask = use_mask
         self.randrange = random.randrange
         self.compress = compress
@@ -536,7 +548,7 @@ class WebSocketWriter:
         self._output_size = 0
         self._compressobj = None
 
-    def _send_frame(self, message, opcode):
+    async def _send_frame(self, message, opcode, compress=None):
         """Send a frame over the websocket with message as its payload."""
         if self._closing:
             ws_logger.warning('websocket connection is closing.')
@@ -546,12 +558,17 @@ class WebSocketWriter:
         # Only compress larger packets (disabled)
         # Does small packet needs to be compressed?
         # if self.compress and opcode < 8 and len(message) > 124:
-        if self.compress and opcode < 8:
-            if not self._compressobj:
-                self._compressobj = zlib.compressobj(wbits=-self.compress)
+        if (compress or self.compress) and opcode < 8:
+            if compress:
+                # Do not set self._compress if compressing is for this frame
+                compressobj = zlib.compressobj(wbits=-compress)
+            else:  # self.compress
+                if not self._compressobj:
+                    self._compressobj = zlib.compressobj(wbits=-self.compress)
+                compressobj = self._compressobj
 
-            message = self._compressobj.compress(message)
-            message = message + self._compressobj.flush(
+            message = compressobj.compress(message)
+            message = message + compressobj.flush(
                 zlib.Z_FULL_FLUSH if self.notakeover else zlib.Z_SYNC_FLUSH)
             if message.endswith(_WS_DEFLATE_TRAILING):
                 message = message[:-4]
@@ -576,146 +593,48 @@ class WebSocketWriter:
             mask = mask.to_bytes(4, 'big')
             message = bytearray(message)
             _websocket_mask(mask, message)
-            self.writer.write(header + mask + message)
+            self.transport.write(header + mask + message)
             self._output_size += len(header) + len(mask) + len(message)
         else:
             if len(message) > MSG_SIZE:
-                self.writer.write(header)
-                self.writer.write(message)
+                self.transport.write(header)
+                self.transport.write(message)
             else:
-                self.writer.write(header + message)
+                self.transport.write(header + message)
 
             self._output_size += len(header) + len(message)
 
         if self._output_size > self._limit:
             self._output_size = 0
-            return self.stream.drain()
+            await self.protocol._drain_helper()
 
-        return noop()
-
-    def pong(self, message=b''):
+    async def pong(self, message=b''):
         """Send pong message."""
         if isinstance(message, str):
             message = message.encode('utf-8')
-        return self._send_frame(message, WSMsgType.PONG)
+        return await self._send_frame(message, WSMsgType.PONG)
 
-    def ping(self, message=b''):
+    async def ping(self, message=b''):
         """Send ping message."""
         if isinstance(message, str):
             message = message.encode('utf-8')
-        return self._send_frame(message, WSMsgType.PING)
+        return await self._send_frame(message, WSMsgType.PING)
 
-    def send(self, message, binary=False):
+    async def send(self, message, binary=False, compress=None):
         """Send a frame over the websocket with message as its payload."""
         if isinstance(message, str):
             message = message.encode('utf-8')
         if binary:
-            return self._send_frame(message, WSMsgType.BINARY)
+            return await self._send_frame(message, WSMsgType.BINARY, compress)
         else:
-            return self._send_frame(message, WSMsgType.TEXT)
+            return await self._send_frame(message, WSMsgType.TEXT, compress)
 
-    def close(self, code=1000, message=b''):
+    async def close(self, code=1000, message=b''):
         """Close the websocket, sending the specified code and message."""
         if isinstance(message, str):
             message = message.encode('utf-8')
         try:
-            return self._send_frame(
+            return await self._send_frame(
                 PACK_CLOSE_CODE(code) + message, opcode=WSMsgType.CLOSE)
         finally:
             self._closing = True
-
-
-def do_handshake(method, headers, stream, protocols=(),
-                 write_buffer_size=DEFAULT_LIMIT, compress=True):
-    """Prepare WebSocket handshake.
-
-    It return HTTP response code, response headers, websocket parser,
-    websocket writer. It does not perform any IO.
-
-    `protocols` is a sequence of known protocols. On successful handshake,
-    the returned response headers contain the first protocol in this list
-    which the server also knows.
-
-    `write_buffer_size` max size of write buffer before `drain()` get called.
-
-    `compress` enable or disable server side deflate extension support.
-    """
-    # WebSocket accepts only GET
-    if method.upper() != hdrs.METH_GET:
-        raise HttpProcessingError(
-            code=405, headers=((hdrs.ALLOW, hdrs.METH_GET),))
-
-    if 'websocket' != headers.get(hdrs.UPGRADE, '').lower().strip():
-        raise HttpBadRequest(
-            message='No WebSocket UPGRADE hdr: {}\n Can '
-            '"Upgrade" only to "WebSocket".'.format(headers.get(hdrs.UPGRADE)))
-
-    if 'upgrade' not in headers.get(hdrs.CONNECTION, '').lower():
-        raise HttpBadRequest(
-            message='No CONNECTION upgrade hdr: {}'.format(
-                headers.get(hdrs.CONNECTION)))
-
-    # find common sub-protocol between client and server
-    protocol = None
-    if hdrs.SEC_WEBSOCKET_PROTOCOL in headers:
-        req_protocols = [str(proto.strip()) for proto in
-                         headers[hdrs.SEC_WEBSOCKET_PROTOCOL].split(',')]
-
-        for proto in req_protocols:
-            if proto in protocols:
-                protocol = proto
-                break
-        else:
-            # No overlap found: Return no protocol as per spec
-            ws_logger.warning(
-                'Client protocols %r don’t overlap server-known ones %r',
-                req_protocols, protocols)
-
-    # check supported version
-    version = headers.get(hdrs.SEC_WEBSOCKET_VERSION, '')
-    if version not in ('13', '8', '7'):
-        raise HttpBadRequest(
-            message='Unsupported version: {}'.format(version),
-            headers=((hdrs.SEC_WEBSOCKET_VERSION, '13'),))
-
-    # check client handshake for validity
-    key = headers.get(hdrs.SEC_WEBSOCKET_KEY)
-    try:
-        if not key or len(base64.b64decode(key)) != 16:
-            raise HttpBadRequest(
-                message='Handshake error: {!r}'.format(key))
-    except binascii.Error:
-        raise HttpBadRequest(
-            message='Handshake error: {!r}'.format(key)) from None
-
-    response_headers = [
-        (hdrs.UPGRADE, 'websocket'),
-        (hdrs.CONNECTION, 'upgrade'),
-        (hdrs.TRANSFER_ENCODING, 'chunked'),
-        (hdrs.SEC_WEBSOCKET_ACCEPT, base64.b64encode(
-            hashlib.sha1(key.encode() + WS_KEY).digest()).decode())]
-
-    notakeover = False
-    if compress:
-        extensions = headers.get(hdrs.SEC_WEBSOCKET_EXTENSIONS)
-        # Server side always get return with no exception.
-        # If something happened, just drop compress extension
-        compress, notakeover = ws_ext_parse(extensions, isserver=True)
-        if compress:
-            enabledext = ws_ext_gen(compress=compress, isserver=True,
-                                    server_notakeover=notakeover)
-            response_headers.append((hdrs.SEC_WEBSOCKET_EXTENSIONS,
-                                     enabledext))
-
-    if protocol:
-        response_headers.append((hdrs.SEC_WEBSOCKET_PROTOCOL, protocol))
-
-    # response code, headers, None, writer, protocol
-    return (101,
-            response_headers,
-            None,
-            WebSocketWriter(
-                stream, limit=write_buffer_size,
-                compress=compress, notakeover=notakeover),
-            protocol,
-            compress)
